#include "litscene.h"
#include "random.h"
#include <math.h>
#include <limits>
#include <fstream>

bool LitScene::load(const char* filename)
{
	std::fstream fin(filename,std::ios::in|std::ios::binary);
	if (fin.is_open())
	{
		// read data
		fin >> *this;
		if (fin.fail())
		{
			std::cerr << "ERROR: LitScene::load, error loading file: " << filename << std::endl;
			fin.close();
			return false;
		}
		fin.close();
		return true;
	}
	else
	{
		std::cerr << "ERROR: LitScene::load, could not open file: " << filename << std::endl;
		return false;
	}
}

ostream& operator<<(ostream& s, LitScene scene)
{
	scene.print(s);
	s << "\n\nLighting:\n";

	int i;
	s << "Ambient: " << scene.Ambient <<"\n";
	for(i=0;i<scene.numberOfLights();++i) s << "Light: " << i << " " << scene.lightAt(i) << "\n";
	for(i=0;i<scene.numberOfAreaLights();++i)
	{
		s << "Area light: " << i << " type: " << scene.areaLightAt(i)->objectType() << " ";
		scene.areaLightAt(i)->print(s);
		s << "\n";
	}

	return s;
}

istream& operator>>(istream& s, LitScene& scene)
//reading
{
	scene.read(s);

	char c=skipWhiteSpace(s);
	if (c!='a')
	{
		cerr << "ERROR: expected ambient code 'a', got unknown code '" << c << "', exiting" << endl;
		exit(1);
	}

	// read the ambient lighting must always be there
	s >> scene.Ambient;

	// read how many lights
	c=skipWhiteSpace(s);
	char d;s>>d;
	if (c!='l'||d!='c')
	{
		cerr << "ERROR: expected light count code 'lc', got unknown code '" << c << d << "', exiting" << endl;
		exit(1);
	}

	int n;
	s >> n;

	if(n==0) return s;

	scene.setLighting(n);

	Light light; int i;		
	for(i=0;i<n;++i)
	{
		c=skipWhiteSpace(s);s>>d;
		if (c!='p'||d!='l')
		{
			cerr << "ERROR: expected point light code 'pl', got unknown code '" << c << d << "', exiting" << endl;
			exit(1);
		}
		s >> light;
		scene.lightAt(i) = light;
	}

	// construct list of area lights
	scene.buildAreaLightList();
	return s;
}

static float power(float x, int n)
{
	if (n < 1) return 1.0;
	if (n == 1) return x;
	else return x * power(x, n - 1);
}

//

Colour LitScene::colourOnObject(GObject *object, Point p, Point eye)
{
	bool shadow = false;
	Vector  n = object->normal(p); //normal vector
	Ray shadowray;

	Colour colour;

	//compute ambient component
	Colour ka = object->material().ambient();
	colour = ka*Ambient;

	for(int i=0; i<NLights; ++i)
	{
		Light light = lightAt(i);
		// Find the direction to the light from the point 
		Vector vL(0,0,0);
		if(light.directional())
		{
			Vector tmpVector = light.vector();
			vL = minus(vL,tmpVector);
			vL = vL.normalised();
		}
		else
		{ 
			Point tmpPoint = light.point();
			vL = (tmpPoint - p);//.normalised();
		}

		/** Your code here (1) - is the light occluded by another object ?**/
		Ray sr;
		sr.origin()=p;
		sr.direction()=vL;

		if (isShadowed(object,sr,light.directional()))
		{ continue; }

		vL = vL.normalised();
		/** does the light illuminate this point? **/
		float nl =n^vL; //dot product
		float nh;
		Vector e = eye-p;
		e.normalise();
		Vector h = add(e,vL);
		h.normalise(); // (V + L) / |L + V|)
		nh = n^h;

		/** if illuminated, add diffuse and specular component **/
		if (nl>0.0f)
		{
			Colour kd = object->material().diffuse();	
			Colour ks = object->material().specular();	
			Colour Intensity = light.intensity();
			Colour Diffuse = (kd * nl);
			Colour Specular = (ks * power(nh, (int) object->material().shininess()));
			Colour AddedColour = (Diffuse+Specular);	
			Colour Reflection = Intensity*AddedColour;
			colour = colour + Reflection;
		}
	}

	colour=colour+object->material().emission();

	//check that the colour is in the bounds 0 to 1
	colour.check();

	return colour;
}

GObject* LitScene::intersect(Ray ray,Point& p,Vector& n)
{
	double tmin =1E+20;
	bool found = false;GObject* object = NULL;
	//for each object
	for (int i=0; i<numObjects(); ++i)
	{
		float t;
		Colour col;
		if (at(i)->intersect(ray,t,col))
		{
			//intersection found
			if ( (t<tmin) && (t>0.0f) )
			{
				object= at(i);
				tmin = t;
				found = true;				
			}
		} 
	}

	if (found)
	{
		p = ray.pointAt((float)tmin);
		n = object->normal(p);
		return object;
	}

	return NULL;
}

bool LitScene::intersect(GObject* me, Ray ray, Colour& colour, int depth)
/**returns true if the ray intersects the scene, and if so then the colour
at the intersection point. This overrides the method in Scene**/
{
	double tmin =1E+20;
	bool found =false;
	GObject* object =NULL;
	Ray reflected;
	if (depth==0) return false; /* end of recursion */

	//for each object
	for (int i=0; i<numObjects(); ++i)
	{
		float t;
		Colour col;
		if ((at(i)!= me) && at(i)->intersect(ray,t,col))
		{
			//intersection found
			if ( (t<tmin) && (t>0.0f) )
			{
				object = at(i);
				tmin = t;
				colour = col;
				found = true;
			}
		} 
	}

	if (found)	// an object has been intersected
	{
		Colour refl_col;
		/*find the intersection point*/
		Point p = ray.pointAt((float)tmin);
		Vector n = object->normal(p); //normal vector

		//now we want the colour computed at point p on the object
		colour = colourOnObject(object, p, ray.origin());
		
		if( object->material().specular().sum() > 0.0) 
		{
			/** Your code here (2) - do a recursive ray trace. Use specular
			colour to modulate the colour returned by the recursive
			intersect call **/
			reflected.origin()=p;
			reflected.direction()=add(n*2.0*(n^(ray.direction().invert())),ray.direction());

			if (intersect(object,reflected,refl_col,depth-1))
			{
				Colour kd = object->material().diffuse();	
				Colour ks = object->material().specular();	
				colour = ks*refl_col + kd*colour;
			}
			else
				colour = Colour(0,0,0);
		}

		//check that the colour is in the bounds [0..1]
		colour.check();  
		return true;
	}
	else 
	{
		colour = Colour(0,0,0);
		return false;
	}
}

bool LitScene::hitLight(GObject* me, Ray ray)
{
	double tmin = 1E+20;
	bool found = false;
	GObject* object = NULL;

	//for each object
	for(int i = 0; i < numObjects(); ++i)
	{
		float t;
		Colour col;
		if((at(i) != me) && at(i)->intersect(ray,t,col)){
			//intersection found
			if((t < tmin) && (t > 0.0) ){
				object = at(i);
				tmin = t;
				found = true;
			}
		} 
	}
	if(found)
	{  
		return object->isEmitter();
	}
	else 
	{
		return false;
	}
}

GObject* LitScene::isShadowed(GObject* me, Ray ray,bool directional)
/*returns true if the ray intersects the scene, and if so then the colour
at the intersection point. This overrides the method in Scene*/
{
	double tmin = 1e+20;
	if (directional) tmin = 1e+20;
	bool found = false;
	GObject* object = NULL;
	Ray reflected;

	//for each object
	for (int i=0; i<numObjects(); ++i)
	{
		float t;
		Colour col;
		if ((at(i)!= me) && at(i)->intersect(ray,t,col))
		{
			//intersection found
			if ( (t<tmin) && (t>0.0) )
			{
				object = at(i);
				tmin = t;
				found = true;
			}
		} 
	}

	if (found) return object;
	
	return NULL;
}

void LitScene::buildAreaLightList(void)
{
	// clear current list
	if (m_areaLights) delete [] m_areaLights;
	m_areaLights=NULL;
	m_areaLightCount=0;

	// count the number of emitting objects
	for (int i=0; i<numObjects(); i++)
	{
		if (at(i))
			if (at(i)->isEmitter())
				m_areaLightCount++;
	}

	// allocate new array
	m_areaLights=new GObject*[m_areaLightCount];
	if (!m_areaLights) exit(1);

	// add emitters to array
	int j=0;
	for (int i=0; i<numObjects(); i++)
	{
		if (at(i))
			if (at(i)->isEmitter())
			{
				m_areaLights[j++]=at(i);
			}
	}
} // buildAreaLightList