#include "debug.h"
#include "assert.h"
#include "brdf.h"
#include "litscene.h"
#include "sphere.h"
#include "polygon.h"
#include "simplecamera.h"
#include <GMmatlib.h>	
#include <math.h>
#include <float.h>


// ===========================================
// PATH TRACING
// ===========================================

Colour LitScene::renderPixel(int i,int j,SimpleCamera& TheCamera,int N_RAYS_PER_PIXEL)
{
	Colour col(0,0,0);

	// ----------------------------------------------------------------------------------
	// This is where you would plug in sampling of N rays (N_RAYS_PER_PIXEL) which would
	// be sampled by stratified/jittered/adaptive sampling
	// ----------------------------------------------------------------------------------

	Ray ray = TheCamera.ray(i,j);	// ray through the centre of pixel (i,j)
	col =tracePath(ray);				// trace this ray & get the colour generated by it

	return col;
}


Colour LitScene::tracePath(const Ray& ray, GObject* object, Colour weightIn, int depth)
{
	float Lamda1 = frand(), Lamda2 = frand();   // random variables in cannonical 2d space in range of [0..1]

	Colour colourOut(0,0,0);

	// ----------------------------
	// check to see wether we should terminate here
	if (depth>=m_maxBounces) return colourOut;
	if (weightIn.sum()<m_cutOffThreshold) return colourOut;

	Point ixp;		// intersection point
	Vector normal;	// normal at intersection point

	// ----------------------------
	// intersect ray with scene
	GObject* hitObject=intersect(ray,ixp,normal);

	// make sure we dont have self intersections & have hit 'something'
	if (!hitObject || hitObject==object) return colourOut;

	if (hitObject->isEmitter())
	{
		// ray struck an emitter - return emission.
		// you could possibly argue that there should be a slight chance that a ray is scattered 
		// off an emitter depending on its brdf - this is however usually ignored.
		float ndotl = normal ^ ray.direction().invert().normalised();
		if (ndotl>0.0f)
			colourOut =hitObject->material().emission();
		return colourOut;
	}
	
	// ----------------------------
	// sample direct illumination
	// ----------------------------
	if (m_areaLightCount>0)
	{
		// choose an emitter randomly
		int theLight=rand()%m_areaLightCount;

		// now sample chosen emitter
		Point lightPoint;
		float pdf;
		if (areaLightAt(theLight)->sample(lightPoint,pdf,ixp, Lamda1,Lamda2))
		{
			// generate ray in sampled direction
			Ray lightray;
			lightray.origin() =ixp;
			lightray.direction() =lightPoint-ixp;

			float ndotl= (normal ^ lightray.direction().normalised());
			if (ndotl>0.0f)
			{
				Vector nlight;
				Point plight;
				if ((areaLightAt(theLight)==intersect(lightray,plight,nlight)))
				{
					float ldist;
					float ndotl_light= lightray.direction().invert().normalised(ldist)^nlight;
					if (ndotl_light>0)
					{						
						Colour brdf= hitObject->brdf()->brdf(ixp,ray.direction(),lightray.direction(),normal, Lamda1);
						
						if ((brdf.sum()*ndotl*(1.0f/pdf)) >3.0f)
							colourOut= colourOut + areaLightAt(theLight)->material().emission()*(ndotl_light/(ldist*ldist));
						else
							colourOut= colourOut + brdf*ndotl*areaLightAt(theLight)->material().emission()*(ndotl_light/(pdf*ldist*ldist));
					}
				}
			}
		}
	}

	// ----------------------------
	// sample indirect illumination
	// ----------------------------

	// As a suggestion, you would need to consider the following:
	// * generate a reflected ray according to BRDF
	// * compute the weight of this ray
	// * recursively trace path
	// * add retreived colour to colourOut

	return colourOut;
}


// ===========================================
// POLYGONAL SAMPLING
// ===========================================
float Polygon::TriangularSampling(Point& p, float s, float t, int Type = 0)
{
	// < Sampling triangular Luminaires > 
	// Reference: (Section 3.3.4) Peter Shirley et al "Monte Carlo Techniques for Direct Lighting Calculations," ACM Transactions on Graphics, 1996

	// Tips:
	// Points in the triangle can be accessed by P[0], P[1] and P[2]
	// The point to sample should be written to p.
	// The probability density function sample should be returned.
	
	return 0.0f;	// nothing for now
}

float Polygon::RectangularSampling(Point& p, float s, float t, int Type = 0)
{
	// < Sampling rectangular Luminaires > 
	// Reference: (Section 3.3.3) Peter Shirley et al "Monte Carlo Techniques for Direct Lighting Calculations," ACM Transactions on Graphics, 1996

	// Tips:
	// Points in the rectangle can be accessed by P[0], P[1], P[2] and P[3]
	// The point to sample should be written to p.
	// The probability density function sample should be returned.

	return 0.0f;	// nothing for now
}

float Polygon::PolygonSampling(Point& p,float s,float t, int Type = 0)
{
	std::cerr << "ERROR: Polygon::PolygonSampling, not implemented, exiting." << std::endl;
	exit(1);
}

bool Polygon::sample(Point& p, float& probability, const Point& from, float s, float t) 
{
	// depending on the number of vertices of polygon, select appropriate sampling scheme
	if (Max==3)			probability = TriangularSampling(p, s, t);
	else if (Max==4)	probability = RectangularSampling(p, s, t);
	else					probability = PolygonSampling(p, s, t);
	return true; 
}


// ===========================================
// SPHERE SAMPLING
// ===========================================

// this samples the sphere over the solid angle as seen from point
// see: Peter Shirley, Changyaw Wang, and Kurt Zimmerman, "Monte Carlo Techniques for Direct Lighting Calculations," ACM Transactions on Graphics, 15(1), 1996, pp. 1-36
bool Sphere::sample(Point& p, float& probability, const Point& from, float s, float t) 
{
	GPVector3 U,V,W;
	GPVector D;
	GPMatrix local2WC;

	// find direction towards sphere in polar coordinates
	float tmp=(*this).radius()/((from-(*this).Centre).norm());
	float tmpSrqt=sqrt(1-(tmp*tmp));

	float costheta=(1-s)+(s*tmpSrqt);
	float theta=acos(costheta);
	float phi=2*(float)PI*t;
	
	// convert to a vector in local coordinates
	D.x=sin(theta)*cos(phi);
	D.y=costheta;
	D.z=sin(theta)*sin(phi);
	D.w=1.0f;

	// the vector is given in a right-handed orthonormal basis [U,V,W], that
	// originates at point, and W is the vector btw point and the centre of the sphere
	Vector v=((*this).Centre-from).normalised();
	V.x=v.x();V.y=v.y();V.z=v.z();
	U=V.orthogonal();
	W=U%V;

	// now form a matrix that transforms the local coordinates to world coordinates
	local2WC.local2WCxForm(U,V,W);
	D=D*local2WC;

	// find the intersection point
	Ray r;r.origin()=from;
	r.direction().x()=D.x;r.direction().y()=D.y;r.direction().z()=D.z;
	float tix; Colour colour;
	if (intersect(r,tix,colour))
	{
		p=r.pointAt(tix);
		Vector n=(p-Centre).normalised();
		Vector d=r.direction().invert().normalised();
		float l=(p-from).norm();
		if (tmpSrqt==1)
			probability=0;
		else
			probability=(d^n)/(2.0f*(float)PI*(l*l)*(1-tmpSrqt));
		return true;
	}
	return false;
}


// ===========================================
// BRDF SAMPLING
// ===========================================
Ray lambertianBRDF::reflection(
							  const Ray& incoming,	// incoming ray
							  const Vector& normal,	// surface normal
							  const Point& point,	// intersection point with surface
							  float s,float t,		// canonical random variables (on unit square)
							  float& pdf
							  ) const
{
	GPMatrix local2WC;
	GPVector3 U,N,V;
	GPVector	D;
	D.w=1.0f;
	
	Ray r;
	r.origin()=point;

	D.x=cos(2.0f*(float)PI*s)*sqrt(1-t);
	D.y=sqrt(t);
	D.z=sin(2.0f*(float)PI*s)*sqrt(1-t);

	N.x=normal.x();
	N.y=normal.y();
	N.z=normal.z();

	// the vector is given in a right-handed orthonormal basis [U,normal,V], form that basis
	U=N.orthogonal();
	V=U%N;

	// now form a matrix that transforms the local coordinates to world coordinates
	local2WC.local2WCxForm(U,N,V);
	D=D*local2WC;
	r.direction().x()=D.x;
	r.direction().y()=D.y;
	r.direction().z()=D.z;

	pdf=(r.direction().normalised()^normal)/(float)PI;
	return r;
}

Colour lambertianBRDF::brdf(
		const Point& x,			// surface point to evaluate brdf at
		const Vector& in,			// incoming vector
		const Vector& out,		// outgoing vector
		const Vector& Nx,			// normal at x (must be unit length)
		const float s
		) const
{
	return m_kd/(float)PI;
}


Ray phongBRDF::reflection(
							  const Ray& incoming,	// incoming ray
							  const Vector& normal,	// surface normal
							  const Point& point,	// intersection point with surface
							  float s,float t,		// canonical random variables (on unit square)							  
							  float& pdf
							  ) const
{
	// Please refer to the references listed for this coursework on how to proceed with this section
	// - does nothing for now...
	Ray r;
	return r;
}

Colour phongBRDF::brdf(
		const Point& x,			// surface point to evaluate brdf at
		const Vector& in,			// incoming vector
		const Vector& out,		// outgoing vector
		const Vector& Nx,			// normal at x (must be unit length)
		const float s
		) const
{
	// Please refer to the references listed for this coursework on how to proceed with this section
	// - does nothing for now...
	return Colour(0,0,0);
}
